from pathlib import Path

import fire
import torch
from torch import nn
from torch.utils.data import DataLoader
from datasets import load_from_disk

from optimus.models.optimus import OptimusTransformer, OptimusConfig
from optimus.trainer import Trainer
from optimus.utils import setup_utils


def main(batch_size: int = 8,
         grad_acc_steps: int = 1,
         seq_len: int = 512,
         lr_max: float = 1e-4,
         grad_clip_norm: float = 1.0,
         epochs: int = 1,
         hf_tokenizer_name: str = 'gpt2',
         data_dir: Path = Path('/workspace'),
         dataset_dir: Path = Path('wikitext103_tokenized_dataset'),
         checkpoints_path: str = 'best_model.pth',
         dim: int = 512,
         n_layers: int = 6,
         n_heads: int = 8,
         dropout: float = 0.0,
         use_fp16: bool = True):
    """
    Run the main training loop for the model.

    Args:
        batch_size (int): Batch size for training.
        grad_acc_steps (int): Number of batches to accumulate gradients for
            before running backpropagation to update weights.
        seq_len (int): Context length for training.
        lr_max (float): Maximum learning rate, used for one-cycle scheduling.
        grad_clip_norm (float): Gradient clipping value for gradient's norm.
        epochs (int): Number of epochs to train for.
        hf_tokenizer_name (str): HuggingFace tokenizer name.
        data_dir (Path): Directory located in a large/fast storage, which holds
            data to be used by the model. Should also be capable to accomodate
            training checkpoints and logging data. This should not be the same
            as the directory of the source code, as this directory can easily
            reach hundreds of gigabytes on big datasets/model sizes.
        dataset_dir (Path): Path to dataset to be used for training. Note:
            Should be tokenized _with the same tokenizer_ as the one used for
            training (`hf_tokenizer_name` above). The dataset should not be
            already split into batches. Given path is appended to `data_dir`.
        checkpoints_path (str): Where to save the trained model. Should be a .pt
            or .pth file.
        dim (int): Dimension of the model.
        n_layers (int): Number of layers for the model.
        n_heads (int): Number of heads inside an attention layer for the model.
        dropout (float): Dropout to use for the model.
        use_fp16 (bool): Whether to train using floating-point 16-bits
            precision.

    """
    if isinstance(data_dir, str):
        data_dir = Path(data_dir)

    if isinstance(dataset_dir, str):
        dataset_dir = data_dir / Path(dataset_dir)
    else:
        dataset_dir = data_dir / dataset_dir

    print(f"Running with:\n"
        f"\t- batch size: {batch_size}\n"
        f"\t- gradient accumulation steps: {grad_acc_steps}\n"
        f"\t- context length: {seq_len}\n"
        f"\t- max learning rate: {lr_max}\n"
        f"\t- gradient clipping norm: {grad_clip_norm}\n"
        f"\t- epochs: {epochs}\n"
        f"\t- huggingface tokenizer: {hf_tokenizer_name}\n"
        f"\t- training data directory: {str(data_dir)}\n"
        f"\t- dataset directory: {str(dataset_dir)}\n"
        f"\t- checkpoints path: {checkpoints_path}\n"
        f"\t- model dimension: {dim}\n"
        f"\t- model layers: {n_layers}\n"
        f"\t- model attention heads: {n_heads}\n"
        f"\t- model dropout: {dropout}\n"
        f"\t- 16-bit floating-point training (fp16): {use_fp16}\n"
        f"Please see '--help' if you want to change these settings")

    # load tokenizer
    tokenizer = setup_utils.create_tokenizer(hf_tokenizer_name)

    # load dataset
    dataset = load_from_disk(str(dataset_dir))

    print(f'Number of examples in training set: {len(dataset['train'])}')
    print(f'Number of examples in testing set: {len(dataset['test'])}')

    # create dataloader objects and move to device
    train_dataloader = DataLoader(
        dataset['train'],
        batch_size=batch_size, # per GPU
        num_workers=6, # allow pre-fetching data through multi-process
        pin_memory=True, # fast CPU-GPU transfer
        shuffle=True,
        drop_last=True,
    )

    eval_dataloader = DataLoader(
        dataset['test'],
        batch_size=batch_size, # per GPU
        num_workers=6, # allow pre-fetching data through multi-process
        pin_memory=True, # fast CPU-GPU transfer
    )

    # create model and move to device
    config = OptimusConfig(vocab_size=len(tokenizer),
                           num_hidden_layers=n_layers,
                           num_attention_heads=n_heads,
                           hidden_size=dim,
                           attention_dropout=dropout,
                           tie_word_embeddings=False)
    model = OptimusTransformer(config)
    model = model.to('cuda')

    _total_params = sum(p.numel() for p in model.parameters())
    print(f'Number of model parameters: {_total_params}')

    # define loss metric
    criterion = nn.CrossEntropyLoss()

    # define optimizer
    # see [1] for a discussion on what the epsilon value should be for amp; 1e-7
    # is a good default for both amp and normal training
    # [1]: https://github.com/pytorch/pytorch/issues/26218
    optimizer = torch.optim.Adam(model.parameters(), betas=(0.9, 0.999), eps=1e-7)

    print('Starting training...')

    # create trainer and start fitting
    trainer = Trainer(
        model=model,
        train_dataloader=train_dataloader,
        eval_dataloader=eval_dataloader,
        tokenizer=tokenizer,
        criterion=criterion,
        optimizer=optimizer,
        lr=lr_max,
        grad_acc_steps=grad_acc_steps,
        grad_clip_norm=grad_clip_norm,
        model_save_path=checkpoints_path,
        use_fp16=use_fp16,
        progress_bar=True
    )
    trainer.fit(epochs)

    print(f"Finished training! Best model weights saved at '{checkpoints_path}'")


if __name__=='__main__':
    fire.Fire(main)
